using System;
using UnityEngine;
using UnityEngine.UI;
using UnityEngine.Events;
using UnityEngine.PostProcessing;
using System.Text;
using UnityEngine.XR;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;

using MVR.FileManagementSecure;
using MeshVR;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using AssetBundles;
using MVR;

namespace ChenRong
{
    public class CanvasButton : MVRScript
    {
        #region PluginInfo    
        public string pluginAuthor = "ChenRong";
        public string pluginName = "CanvasButton";
        public string pluginVersion = "1.0";
        public string pluginDate = "2024-01-18";
        public string pluginDescription = @"Canvas Button Plugin";
        #endregion

        private bool active = true;
        public static Canvas canvas = null;

        private InputField action_input = null;
        private List<string> button_names = new List<string>();
        private Dictionary<string, EventTrigger> button_events = new Dictionary<string, EventTrigger>();
        private Dictionary<string, UIDynamicButton> button_handles = new Dictionary<string, UIDynamicButton>();

        public override void Init()
        {
            JSONStorableString helpText = new JSONStorableString(
                "About",
                "<color=#000><size=35><b>CanvasButton</b></size></color>\n"
                    + "<color=#333>"
                    + "Add Button to 2D UI interface\n"
                    + "</color>"
            );
            UIDynamic helpTextfield = CreateTextField(helpText, true);
            helpTextfield.height = 120.0f;

            JSONStorableAction toggle_action = new JSONStorableAction("Toggle", OnToggleButton);
            RegisterAction(toggle_action);
            JSONStorableAction show_action = new JSONStorableAction("Show", OnShowButton);
            RegisterAction(show_action);
            JSONStorableAction hide_action = new JSONStorableAction("Hide", OnHideButton);
            RegisterAction(hide_action);
        }


        public void Start()
        {
            DMEventHandler.LoadAssets();

            float worldScale = SuperController.singleton.worldScale;
            SuperController.singleton.worldScale = 1.0f;
            create_buttons_canvas();
            SuperController.singleton.worldScale = worldScale;

            try
            {
                string action_name = "Action 1";
                if (button_names.Count > 0)
                    action_name = create_action_name(button_names[button_names.Count - 1]);

                JSONStorableString positionTitle = new JSONStorableString("ActionName", "0");
                UIDynamicTextField tmp2Textfield = CreateTextField(positionTitle);
                setup_text_field(tmp2Textfield, 50f, false, false);
                action_input = tmp2Textfield.UItext.gameObject.AddComponent<InputField>();
                action_input.textComponent = tmp2Textfield.UItext;
                action_input.lineType = InputField.LineType.SingleLine;
                action_input.text = action_name;
                action_input.onValueChanged.AddListener(delegate { OnButtonNameTextChanged(); });

                var btn = CreateButton("Add Button", false);
                btn.button.onClick.AddListener(() =>
                {
                    OnAddButton();
                });

                btn = CreateButton("Toggle Button", false);
                btn.button.onClick.AddListener(() =>
                {
                    OnToggleButton();
                });

                JSONStorableString buttonText = new JSONStorableString(
                    "Button Message",
                    "<color=#000><size=45><b>Button Information</b></size></color>"
                );
                UIDynamic buttonTextfield = CreateTextField(buttonText);
                buttonTextfield.height = 20.0f;

                for (int i = 0; i < button_names.Count; i++)
                {
                    var name = button_names[i];

                    btn = CreateButton(name + " Trigger", false);
                    btn.button.onClick.AddListener(() =>
                    {
                        button_events[name].OpenPanelActionStart();
                    });

                    button_handles.Add(name, add_button(
                        name,
                        () =>
                        {
                            button_events[name].Trigger();
                        },
                        0,
                        i
                    ));
                }

            }
            catch (Exception e)
            {
                SuperController.LogError("start: " + e);
                DestroyImmediate(this);
            }
        }
        void Update()
        {
            if (SuperController.singleton.isLoading)
                return;
        }
        void FixedUpdate() { }
        void OnDestroy()
        {
            try
            {
                if (SuperController.singleton != null)
                    SuperController.singleton.RemoveCanvas(canvas);

                if (canvas != null)
                {
                    canvas.transform.SetParent(null, false);
                    if (canvas.gameObject != null)
                        GameObject.Destroy(canvas.gameObject);
                }
            }
            catch (Exception e)
            {
                SuperController.LogError("Exception caught: " + e);
            }
        }

        private void OnButtonNameTextChanged()
        {
            // SuperController.LogMessage(">> action_input " + action_input.text);
        }

        private void OnAddButton()
        {
            var name = action_input.text;

            button_names.Add(name);
            button_events.Add(name, new EventTrigger(this, name));
            button_handles.Add(name, add_button(
                name,
                () =>
                {
                    button_events[name].Trigger();
                },
                0,
                button_names.Count - 1
            ));

            var btn = CreateButton(name + " Trigger", false);
            btn.button.onClick.AddListener(() =>
            {
                button_events[name].OpenPanelActionStart();
            });
            button_events[name].OpenPanelActionStart();

            action_input.text = create_action_name(name);
        }



        private void OnShowButton()
        {
            active = true;
            set_active(active);
        }
        private void OnHideButton()
        {
            active = false;
            set_active(active);
        }

        private void OnToggleButton()
        {
            active = !active;
            set_active(active);
        }
        private void setup_text_field(UIDynamicTextField target, float fieldHeight, bool disableBackground = true, bool disableScroll = true)
        {
            if (disableBackground) target.backgroundColor = new Color(1f, 1f, 1f, 0f);
            LayoutElement tfLayout = target.GetComponent<LayoutElement>();
            tfLayout.preferredHeight = tfLayout.minHeight = fieldHeight;
            target.height = fieldHeight;
        }
        private void set_active(bool active)
        {
            foreach (var name in button_names) button_handles[name].gameObject.SetActive(active);
        }
        private string create_action_name(string name)
        {
            var str = System.Text.RegularExpressions.Regex.Replace(name, @"[^0-9]+", "");
            if (str.Length > 0)
            {
                var no = int.Parse(str);
                var newn = name.Substring(0, name.Length - str.Length) + (no + 1).ToString();
                return newn;
            }
            else
                return name + " 1";
        }
        public void create_buttons_canvas()
        {
            try
            {
                clear_canvas();

                GameObject canvasObject = new GameObject();
                canvas = canvasObject.AddComponent<Canvas>();
                canvas.renderMode = RenderMode.WorldSpace;
                canvas.pixelPerfect = false;
                SuperController.singleton.AddCanvas(canvas);

                canvas.transform.SetParent(SuperController.singleton.mainHUD, false);

                CanvasScaler cs = canvasObject.AddComponent<CanvasScaler>();
                cs.scaleFactor = 80.0f;
                cs.dynamicPixelsPerUnit = 1f;

                GraphicRaycaster gr = canvasObject.AddComponent<GraphicRaycaster>();

                float scale = 0.001f;
                canvas.transform.localScale = new Vector3(scale, scale, scale);

                /* set position */
                // canvas.transform.localPosition = new Vector3(-0.7f, 0, 0);
                // canvas.transform.localPosition = new Vector3(0.5f, -0.66f, 0.35f);
                canvas.transform.localPosition = new Vector3(-1.9f, 0.45f, 0f);
                // canvas.transform.localPosition = new Vector3(-1.9f, -0.955f, 0f);

                /* look at camera */
                canvas.transform.localEulerAngles = new Vector3(1, 180, 0);
                canvas.transform.Translate(0, 0.2f, 0);
            }
            catch (Exception e)
            {
                SuperController.LogError("create button failed: " + e);
                DestroyImmediate(this);
            }
        }


        public void clear_canvas()
        {
            try
            {
                if (canvas != null)
                {
                    if (SuperController.singleton != null)
                        SuperController.singleton.RemoveCanvas(canvas);

                    canvas.transform.SetParent(null, false);

                    if (canvas.gameObject != null)
                        GameObject.Destroy(canvas.gameObject);
                }
            }
            catch (Exception e)
            {
                SuperController.LogError("clear_canvas failed: " + e);
                DestroyImmediate(this);
            }
        }

        public UIDynamicButton add_button(
            string name,
            UnityAction callback,
            int column,
            int row,
            int width = 40,
            int height = 40
        )
        {
            float xSpacing = 0.16f; // 0.22f
            float ySpacing = 0.05f; // 0.05f;
            Color accessButtonColor = new Color(0.8392f, 0.8392f, 0.8392f);
            Color accessTextColor = new Color(0, 0, 0);

            UIDynamicButton button = create_button(name, width, height);

            button.button.onClick.AddListener(callback);

            RegisterAction(new JSONStorableAction(name, () =>
                {
                    button_events[name].Trigger();
                }));

            button.transform.Translate(
                -column * xSpacing,
                0.375f + -row * ySpacing,
                0.33f,
                Space.Self
            );

            color_button(button, accessTextColor, accessButtonColor);

            return button;
        }

        public UIDynamicButton create_button(string name, float width = 100, float height = 80)
        {
            Transform button = GameObject.Instantiate<Transform>(
                this.manager.configurableButtonPrefab
            );
            ConfigureTransform(button, width, height);
            ParentToCanvas(button);

            UIDynamicButton uiButton = button.GetComponent<UIDynamicButton>();
            uiButton.label = name;
            uiButton.buttonText.fontSize = 20;
            return uiButton;
        }

        public static void color_button(UIDynamicButton button, Color textColor, Color buttonColor)
        {
            button.textColor = textColor;
            button.buttonColor = buttonColor;
        }

        private void ConfigureTransform(Transform t, float width, float height)
        {
            t.transform.position = Vector3.zero;
            RectTransform rt = t.GetComponent<RectTransform>();
            rt.anchoredPosition = new Vector2(width / 2, height / 2);
            rt.SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, width);
            rt.SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, height);
        }

        private void ParentToCanvas(Transform t)
        {
            t.SetParent(canvas.transform, false);
        }

        public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
        {
            JSONClass jc = base.GetJSON(includePhysical, includeAppearance, forceStore);
            if (includePhysical || forceStore)
            {
                needsStore = true;

                jc["js_button_counts"] = button_names.Count.ToString();

                for (int i = 0; i < button_names.Count; i++)
                {
                    jc["js_button_name_" + i] = button_names[i];
                    jc[button_events[button_names[i]].Name] = button_events[button_names[i]].GetJSON(base.subScenePrefix);

                    // SuperController.LogMessage(">> 1 GetJSON " + button_names[i].Name + " - " + base.subScenePrefix);
                }
            }
            return jc;
        }

        public override void RestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, JSONArray presetAtoms = null, bool setMissingToDefault = true)
        {
            base.RestoreFromJSON(jc, restorePhysical, restoreAppearance, presetAtoms, setMissingToDefault);
            // SuperController.LogMessage(">> 1 RestoreFromJSON " + jc.ToString());

            if (jc.HasKey("js_button_counts"))
            {
                for (int i = 0; i < int.Parse(jc["js_button_counts"]); i++)
                {
                    if (jc.HasKey("js_button_name_" + i))
                    {
                        string button_name = jc["js_button_name_" + i];
                        button_names.Add(button_name);
                        button_events.Add(button_name, new EventTrigger(this, button_name));
                    }
                }
            }

            base.RestoreFromJSON(jc, restorePhysical, restoreAppearance, presetAtoms, setMissingToDefault);
        }

        public override void LateRestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
        {
            base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);

            if (!base.physicalLocked && restorePhysical && !IsCustomPhysicalParamLocked("trigger"))
            {
                if (jc.HasKey("js_button_counts"))
                {
                    for (int i = 0; i < int.Parse(jc["js_button_counts"]); i++)
                    {
                        if (jc.HasKey("js_button_name_" + i))
                        {
                            string button_name = jc["js_button_name_" + i];
                            button_events[button_name].RestoreFromJSON(jc, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
                        }
                    }
                }
            }
        }
    }


    public class EventTrigger : Trigger
    {
        public MVRScript Owner
        {
            get; private set;
        }

        private string _name;

        public string Name
        {
            get { return _name; }
        }

        public EventTrigger(MVRScript owner, string name)
        {
            Owner = owner;
            _name = name;
            handler = DMEventHandler.Instance;
        }

        public void OpenPanelActionStart()
        {
            if (!DMEventHandler.Loaded)
            {
                SuperController.LogError("EventTrigger: You need to call DMEventHandler.LoadAssets() before use.");
                return;
            }

            triggerActionsParent = Owner.UITransform;
            InitTriggerUI();
            OpenTriggerActionsPanel();

            triggerActionsPanel.Find("Panel/Header Text").GetComponent<Text>().text = _name;
            triggerActionsPanel.Find("Content/Tab1/Label").GetComponent<Text>().text = "Event Actions";
            triggerActionsPanel.Find("Content/Tab2").gameObject.SetActive(false);
            triggerActionsPanel.Find("Content/Tab3").gameObject.SetActive(false);
        }

        public void RestoreFromJSON(JSONClass jc, string subScenePrefix, bool isMerge, bool setMissingToDefault)
        {
            if (jc.HasKey(_name))
            {
                JSONClass tc = jc[_name].AsObject;
                if (tc != null)
                    base.RestoreFromJSON(tc, subScenePrefix, isMerge);
            }
            else if (setMissingToDefault)
            {
                base.RestoreFromJSON(new JSONClass());
            }
        }

        public void Trigger()
        {
            active = true;
            active = false;
        }
    }

    public class DMEventHandler : TriggerHandler
    {
        public static bool Loaded { get; private set; }

        private static DMEventHandler myInstance;

        private RectTransform myTriggerActionsPrefab;
        private RectTransform myTriggerActionMiniPrefab;
        private RectTransform myTriggerActionDiscretePrefab;
        private RectTransform myTriggerActionTransitionPrefab;

        public static DMEventHandler Instance
        {
            get
            {
                if (myInstance == null)
                    myInstance = new DMEventHandler();
                return myInstance;
            }
        }

        public static void LoadAssets()
        {
            SuperController.singleton.StartCoroutine(Instance.LoadAssetsInternal());
        }

        private IEnumerator LoadAssetsInternal()
        {
            foreach (var x in LoadAsset("z_ui2", "TriggerActionsPanel", p => myTriggerActionsPrefab = p))
                yield return x;
            foreach (var x in LoadAsset("z_ui2", "TriggerActionMiniPanel", p => myTriggerActionMiniPrefab = p))
                yield return x;
            foreach (var x in LoadAsset("z_ui2", "TriggerActionDiscretePanel", p => myTriggerActionDiscretePrefab = p))
                yield return x;
            foreach (var x in LoadAsset("z_ui2", "TriggerActionTransitionPanel", p => myTriggerActionTransitionPrefab = p))
                yield return x;

            Loaded = true;
        }

        private IEnumerable LoadAsset(string assetBundleName, string assetName, Action<RectTransform> assign)
        {
            AssetBundleLoadAssetOperation request = AssetBundleManager.LoadAssetAsync(assetBundleName, assetName, typeof(GameObject));
            if (request == null)
                throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null request.");
            yield return request;
            GameObject go = request.GetAsset<GameObject>();
            if (go == null)
                throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null GameObject.");
            RectTransform prefab = go.GetComponent<RectTransform>();
            if (prefab == null)
                throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null RectTansform.");
            assign(prefab);
        }


        void TriggerHandler.RemoveTrigger(Trigger t)
        {
            // nothing to do
        }

        void TriggerHandler.DuplicateTrigger(Trigger t)
        {
            throw new NotImplementedException();
        }

        RectTransform TriggerHandler.CreateTriggerActionsUI()
        {
            return UnityEngine.Object.Instantiate(myTriggerActionsPrefab);
        }

        RectTransform TriggerHandler.CreateTriggerActionMiniUI()
        {
            return UnityEngine.Object.Instantiate(myTriggerActionMiniPrefab);
        }

        RectTransform TriggerHandler.CreateTriggerActionDiscreteUI()
        {
            return UnityEngine.Object.Instantiate(myTriggerActionDiscretePrefab);
        }

        RectTransform TriggerHandler.CreateTriggerActionTransitionUI()
        {
            return UnityEngine.Object.Instantiate(myTriggerActionTransitionPrefab);
        }

        void TriggerHandler.RemoveTriggerActionUI(RectTransform rt)
        {
            UnityEngine.Object.Destroy(rt?.gameObject);
        }
    }

}
